"""."""

from ..pwrsupply.psctrl.pscstatus import PSCStatus as _PSCStatus

from .device import Device as _Device


class EGBias(_Device):
    """EGun Bias Device."""

    PWRSTATE = _PSCStatus.PWRSTATE

    class DEVICES:
        """Devices names."""

        LI = 'LI-01:EG-BiasPS'
        ALL = (LI, )

    _properties = (
        'voltoutsoft', 'voltinsoft', 'currentinsoft', 'switch', 'swstatus')

    def __init__(self, devname=None):
        """."""
        if devname is None:
            devname = EGBias.DEVICES.LI

        # check if device exists
        if devname not in EGBias.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        super().__init__(devname, properties=EGBias._properties)

    @property
    def voltage(self):
        """."""
        return self['voltinsoft']

    @voltage.setter
    def voltage(self, value):
        """."""
        self['voltoutsoft'] = value

    @property
    def current(self):
        """."""
        return self['currentinsoft']

    def cmd_turn_on(self):
        """."""
        self['switch'] = self.PWRSTATE.On

    def cmd_turn_off(self):
        """."""
        self['switch'] = self.PWRSTATE.Off

    def is_on(self):
        """."""
        return self['swstatus'] == self.PWRSTATE.On


class EGFilament(_Device):
    """EGun Filament Device."""

    PWRSTATE = _PSCStatus.PWRSTATE

    class DEVICES:
        """Devices names."""

        LI = 'LI-01:EG-FilaPS'
        ALL = (LI, )

    _properties = (
        'voltinsoft', 'currentinsoft', 'currentoutsoft', 'switch', 'swstatus')

    def __init__(self, devname=None):
        """."""
        if devname is None:
            devname = EGFilament.DEVICES.LI

        # check if device exists
        if devname not in EGFilament.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        super().__init__(devname, properties=EGFilament._properties)

    @property
    def voltage(self):
        """."""
        return self['voltinsoft']

    @property
    def current(self):
        """."""
        return self['currentinsoft']

    @current.setter
    def current(self, value):
        """."""
        self['currentoutsoft'] = value

    def cmd_turn_on(self):
        """."""
        self['switch'] = self.PWRSTATE.On

    def cmd_turn_off(self):
        """."""
        self['switch'] = self.PWRSTATE.Off

    def is_on(self):
        """."""
        return self['swstatus'] == self.PWRSTATE.On


class EGHVPS(_Device):
    """Egun High-Voltage Power Supply Device."""

    PWRSTATE = _PSCStatus.PWRSTATE

    class DEVICES:
        """Devices names."""

        LI = 'LI-01:EG-HVPS'
        ALL = (LI, )

    _properties = (
        'currentinsoft', 'currentoutsoft',
        'voltinsoft', 'voltoutsoft',
        'enable', 'enstatus',
        'switch', 'swstatus')

    def __init__(self, devname=None):
        """."""
        if devname is None:
            devname = EGHVPS.DEVICES.LI

        # check if device exists
        if devname not in EGHVPS.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        super().__init__(devname, properties=EGHVPS._properties)

    @property
    def current(self):
        """."""
        return self['currentinsoft']

    @current.setter
    def current(self, value):
        """."""
        self['currentoutsoft'] = value

    @property
    def voltage(self):
        """."""
        return self['voltinsoft']

    @voltage.setter
    def voltage(self, value):
        self['voltoutsoft'] = value

    def cmd_turn_on(self):
        """."""
        self['enable'] = self.PWRSTATE.On
        self['switch'] = self.PWRSTATE.On

    def cmd_turn_off(self):
        """."""
        self['enable'] = self.PWRSTATE.Off
        self['switch'] = self.PWRSTATE.Off

    def is_on(self):
        """."""
        ison = self['enstatus'] == self.PWRSTATE.On
        ison &= self['swstatus'] == self.PWRSTATE.On
        return ison



"""."""

import numpy as _np

from .device import Device as _Device, ProptyDevice as _ProptyDevice


class EVG(_Device):
    """."""

    DEVNAME = 'AS-RaMO:TI-EVG'

    _properties = (
        'InjectionEvt-Sel', 'InjectionEvt-Sts', 'UpdateEvt-Cmd',
        'ContinuousEvt-Sel', 'ContinuousEvt-Sts',
        'RepeatBucketList-SP', 'RepeatBucketList-RB',
        'BucketList-SP', 'BucketList-RB', 'BucketList-Mon',
        'BucketListLen-Mon', 'TotalInjCount-Mon', 'InjCount-Mon',
        )

    def __init__(self):
        """."""
        super().__init__(EVG.DEVNAME, properties=EVG._properties)

    @property
    def nrpulses(self):
        """."""
        return self['RepeatBucketList-RB']

    @nrpulses.setter
    def nrpulses(self, value):
        """."""
        self['RepeatBucketList-SP'] = bool(value)

    @property
    def bucketlist_len(self):
        """."""
        return self['BucketListLen-Mon']

    @property
    def bucketlist_mon(self):
        """."""
        return self['BucketList-Mon']

    @property
    def bucketlist(self):
        """."""
        return self['BucketList-RB']

    @bucketlist.setter
    def bucketlist(self, value):
        """."""
        self['BucketList-SP'] = _np.array(value, dtype=int)

    @property
    def continuous_state(self):
        """."""
        return self['ContinuousEvt-Sts']

    @continuous_state.setter
    def continuous_state(self, value):
        """."""
        self['ContinuousEvt-Sel'] = bool(value)

    @property
    def injection_state(self):
        """."""
        return self['InjectionEvt-Sts']

    @injection_state.setter
    def injection_state(self, value):
        """."""
        self['InjectionEvt-Sel'] = bool(value)

    @property
    def injection_count_total(self):
        """."""
        return self['TotalInjCount-Mon']

    @property
    def injection_count(self):
        """."""
        return self['InjCount-Mon']

    def fill_bucketlist(self, stop, start=1, step=30):
        """."""
        self.bucketlist = _np.arange(start=start, stop=stop, step=step)

    def wait_injection_finish(self, timeout=10):
        """."""
        return self._wait(propty='InjectionEvt-Sts', value=0, timeout=timeout)

    def cmd_update_events(self):
        """."""
        self['UpdateEvt-Cmd'] = 1

    def cmd_turn_on_injection(self, timeout=10):
        """."""
        self.injection_state = 1
        self._wait(propty='InjectionEvt-Sel', value=1, timeout=timeout)

    def cmd_turn_off_injection(self, timeout=10):
        """."""
        self.injection_state = 0
        self._wait(propty='InjectionEvt-Sel', value=0, timeout=timeout)

    def cmd_turn_on_continuous(self, timeout=10):
        """."""
        self.continuous_state = 1
        self._wait(propty='ContinuousEvt-Sel', value=1, timeout=timeout)

    def cmd_turn_off_continuous(self, timeout=10):
        """."""
        self.continuous_state = 0
        self._wait(propty='ContinuousEvt-Sel', value=0, timeout=timeout)


class Event(_ProptyDevice):
    """."""

    _properties = (
        'Delay-SP', 'Delay-RB', 'DelayRaw-SP', 'DelayRaw-RB',
        'DelayType-Sel', 'DelayType-Sts', 'Mode-Sel', 'Mode-Sts',
        'Code-Mon', 'ExtTrig-Cmd',
        )

    MODES = ('Disable', 'Continuous', 'Injection', 'OneShot', 'External')
    DELAYTYPES = ('Incr', 'Fixed')

    def __init__(self, evtname):
        """."""
        super().__init__(
            EVG.DEVNAME, evtname, properties=Event._properties)

    @property
    def mode(self):
        """."""
        return self['Mode-Sts']

    @mode.setter
    def mode(self, value):
        self._enum_setter('Mode-Sel', value, Event.MODES)

    @property
    def mode_str(self):
        """."""
        return Event.MODES[self['Mode-Sts']]

    @property
    def code(self):
        """."""
        return self['Code-Mon']

    @property
    def delay_type(self):
        """."""
        return self['DelayType-Sts']

    @delay_type.setter
    def delay_type(self, value):
        self._enum_setter('DelayType-Sel', value, Event.DELAYTYPES)

    @property
    def delay_type_str(self):
        """."""
        return Event.DELAYTYPES[self['DelayType-Sts']]

    @property
    def delay(self):
        """."""
        return self['Delay-RB']

    @delay.setter
    def delay(self, value):
        self['Delay-SP'] = value

    @property
    def delay_raw(self):
        """."""
        return self['DelayRaw-RB']

    @delay_raw.setter
    def delay_raw(self, value):
        self['DelayRaw-SP'] = int(value)

    def cmd_external_trigger(self):
        """."""
        self['ExtTrig-Cmd'] = 1

    @property
    def is_in_inj_table(self):
        """."""
        return self.mode_str in Event.MODES[1:4]



#!/usr/local/bin/python-sirius
"""Driver module."""

from pcaspy import Alarm as _Alarm, Severity as _Severity

from ...namesys import SiriusPVName
from ..app import App as _App
from ..pvs import ComputedPV as _ComputedPV

from .csdev import conv_dev_2_liname, Const as _Const
from .pvs import LIScalarDiffPV as _LIScalarDiffPV, \
    LIVecDiffPV as _LIVecDiffPV, \
    LIRFStatusPV as _LIRFStatusPV, \
    LIEGHVStatusPV as _LIEGHVStatusPV, \
    LIFilaPSStatusPV as _LIFilaPSStatusPV


class LIDiagApp(_App):
    """Main application responsible for updating DB."""

    def _create_computed_pvs(self, *args):
        # # RF devices
        for dev in _Const.RF_DEVICES:
            devname = SiriusPVName(self._prefix + dev)
            liname = conv_dev_2_liname(dev)

            # DiagAmpDiff-Mon
            pvs = [None]*2
            pvs[_LIScalarDiffPV.SP] = liname + ':SET_AMP'
            pvs[_LIScalarDiffPV.RB] = liname + ':GET_AMP'
            pvo = _ComputedPV(
                devname + ':DiagAmpDiff-Mon', _LIScalarDiffPV(),
                self._queue, pvs, monitor=False)
            self.pvs.append(pvo)

            # DiagPhaseDiff-Mon
            pvs = [None]*2
            pvs[_LIScalarDiffPV.SP] = liname + ':SET_PHASE'
            pvs[_LIScalarDiffPV.RB] = liname + ':GET_PHASE'
            pvo = _ComputedPV(
                devname + ':DiagPhaseDiff-Mon', _LIScalarDiffPV(),
                self._queue, pvs, monitor=False)
            self.pvs.append(pvo)

            # DiagIxQDiff-Mon
            pvs = [None]*4
            pvs[_LIVecDiffPV.I_SETT] = liname + ':GET_CH1_SETTING_I'
            pvs[_LIVecDiffPV.Q_SETT] = liname + ':GET_CH1_SETTING_Q'
            pvs[_LIVecDiffPV.I_DATA] = liname + ':GET_CH1_I'
            pvs[_LIVecDiffPV.Q_DATA] = liname + ':GET_CH1_Q'
            pvo = _ComputedPV(
                devname + ':DiagIxQDiff-Mon', _LIVecDiffPV(),
                self._queue, pvs, monitor=False)
            self.pvs.append(pvo)

            # DiagStatus-Mon
            pvs = [None]*7
            pvs[_LIRFStatusPV.PV_STREAM] = liname + ':GET_STREAM'
            pvs[_LIRFStatusPV.PV_EXTTRG] = \
                liname + ':GET_EXTERNAL_TRIGGER_ENABLE'
            pvs[_LIRFStatusPV.PV_INTEGR] = liname + ':GET_INTEGRAL_ENABLE'
            pvs[_LIRFStatusPV.PV_FBMODE] = liname + ':GET_FB_MODE'
            pvs[_LIRFStatusPV.PV_AMPDIF] = devname + ':DiagAmpDiff-Mon'
            pvs[_LIRFStatusPV.PV_PHSDIF] = devname + ':DiagPhaseDiff-Mon'
            pvs[_LIRFStatusPV.PV_IXQDIF] = devname + ':DiagIxQDiff-Mon'
            pvo = _ComputedPV(
                devname + ':DiagStatus-Mon', _LIRFStatusPV(),
                self._queue, pvs, monitor=False)
            self.pvs.append(pvo)

        # # Egun devices
        # HVPS
        devname = SiriusPVName(self._prefix + _Const.HVPS)
        liname = conv_dev_2_liname(devname)
        # DiagVoltDiff-Mon
        pvs = [None]*2
        pvs[_LIScalarDiffPV.SP] = liname + ':voltoutsoft'
        pvs[_LIScalarDiffPV.RB] = liname + ':voltinsoft'
        pvo = _ComputedPV(
            devname + ':DiagVoltDiff-Mon', _LIScalarDiffPV(),
            self._queue, pvs, monitor=False)
        self.pvs.append(pvo)
        # DiagStatus-Mon
        pvs = [None]*3
        pvs[_LIEGHVStatusPV.PV_SWITCH] = liname + ':swstatus'
        pvs[_LIEGHVStatusPV.PV_ENABLE] = liname + ':enstatus'
        pvs[_LIEGHVStatusPV.PV_VLTDIF] = devname + ':DiagVoltDiff-Mon'
        pvo = _ComputedPV(
            devname + ':DiagStatus-Mon', _LIEGHVStatusPV(),
            self._queue, pvs, monitor=False)
        self.pvs.append(pvo)

        # FilaPS
        devname = SiriusPVName(self._prefix + _Const.FILA)
        liname = conv_dev_2_liname(devname)
        # DiagCurrentDiff-Mon
        pvs = [None]*2
        pvs[_LIScalarDiffPV.SP] = liname + ':currentoutsoft'
        pvs[_LIScalarDiffPV.RB] = liname + ':currentinsoft'
        pvo = _ComputedPV(
            devname + ':DiagCurrentDiff-Mon', _LIScalarDiffPV(),
            self._queue, pvs, monitor=False)
        self.pvs.append(pvo)
        # DiagStatus-Mon
        pvs = [None]*2
        pvs[_LIFilaPSStatusPV.PV_SWITCH] = liname + ':swstatus'
        pvs[_LIFilaPSStatusPV.PV_CURDIF] = devname + ':DiagCurrentDiff-Mon'
        pvo = _ComputedPV(
            devname + ':DiagStatus-Mon', _LIFilaPSStatusPV(),
            self._queue, pvs, monitor=False)
        self.pvs.append(pvo)

        self._pvs_connected = {pv: False for pv in self.pvs}

    def _update_pvs(self):
        for pvo in self.pvs:
            if not pvo.connected:
                if self._pvs_connected[pvo]:
                    self.run_callbacks(
                        pvo.pvname, alarm=_Alarm.TIMEOUT_ALARM,
                        severity=_Severity.INVALID_ALARM,
                        field='status')
                self._pvs_connected[pvo] = False
                if 'DiagStatus' in pvo.pvname:
                    self.run_callbacks(pvo.pvname, value=pvo.value)
            else:
                if not self._pvs_connected[pvo]:
                    self.run_callbacks(
                        pvo.pvname, alarm=_Alarm.NO_ALARM,
                        severity=_Severity.NO_ALARM, field='status')
                self._pvs_connected[pvo] = True
                self.run_callbacks(pvo.pvname, value=pvo.value)



"""."""

import time as _time

from .device import DeviceNC as _DeviceNC


class LILLRF(_DeviceNC):
    """."""

    class DEVICES:
        """Devices names."""

        LI_SHB = 'LA-RF:LLRF:BUN1'
        LI_KLY1 = 'LA-RF:LLRF:KLY1'
        LI_KLY2 = 'LA-RF:LLRF:KLY2'
        ALL = (LI_SHB, LI_KLY1, LI_KLY2)

    _properties = (
        'SET_AMP', 'GET_AMP',
        'SET_PHASE', 'GET_PHASE')

    def __init__(self, devname):
        """."""
        # check if device exists
        if devname not in LILLRF.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        super().__init__(devname, properties=LILLRF._properties)

    @property
    def amplitude(self):
        """."""
        return self['GET_AMP']

    @amplitude.setter
    def amplitude(self, value):
        self['SET_AMP'] = value

    @property
    def phase(self):
        """."""
        return self['GET_PHASE']

    @phase.setter
    def phase(self, value):
        self['SET_PHASE'] = value

    def cmd_set_phase(self, value, timeout=10):
        """."""
        self.phase = value
        self._wait_rb_sp(timeout, 'phase')

    def cmd_set_amplitude(self, value, timeout=30):
        """."""
        self.amplitude = value
        self._wait_rb_sp(timeout, 'amplitude')

    def _wait_rb_sp(self, timeout=10, propty=None):
        """."""
        nrp = int(timeout / 0.1)
        for _ in range(nrp):
            _time.sleep(0.1)
            if propty == 'phase':
                if abs(self.phase - self['SET_PHASE']) < 0.1:
                    break
            elif propty == 'amplitude':
                if abs(self.amplitude - self['SET_AMP']) < 0.1:
                    break
            else:
                raise Exception(
                    'Set LLRF property (phase or amplitude)')
        else:
            print('timed out waiting LLRF.')



"""SI Bunch-by-bunch feedback system configuration."""
from copy import deepcopy as _dcopy
import numpy as _np

_ZERO_INT = 0
_ZERO_FLOAT = 0.0
_EMPTY_NDARRAY_INT = _np.zeros(32)
_EMPTY_STR = ''


def get_dict():
    """Return configuration type dictionary."""
    module_name = __name__.split('.')[-1]
    _dict = {
        'config_type_name': module_name,
        'value': _dcopy(_template_dict),
        'check': False,
    }
    return _dict


# When using this type of configuration to set the machine,
# the list of PVs should be processed in the same order they are stored
# in the configuration. The second numeric parameter in the pair is the
# delay [s] the client should wait before setting the next PV.


_pvs_bbb_proc_h = [
    ['SI-Glob:DI-BbBProc-H:GDEN', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:FBCTRL', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:SETSEL', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:DELAY', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:CLKRST', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:SHIFTGAIN', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:PROC_DS', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:COEFF', _EMPTY_NDARRAY_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:CSET0', _EMPTY_NDARRAY_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:CSET1', _EMPTY_NDARRAY_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:LDSET', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:DESC_COEFF', _EMPTY_STR, 0.0],
    ['SI-Glob:DI-BbBProc-H:DESC_CSET0', _EMPTY_STR, 0.0],
    ['SI-Glob:DI-BbBProc-H:DESC_CSET1', _EMPTY_STR, 0.0],
    ['SI-Glob:DI-BbBProc-H:FID_DELAY', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:FLT_GAIN', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:FLT_PHASE', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:FLT_FREQ', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:FLT_TAPS', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:FTF_TUNE', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DATA31', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DATA30', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DATA29', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DATA28', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DATA27', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DATA26', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DATA25', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DATA24', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DATA23', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DATA22', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DATA21', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DATA20', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DATA19', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DATA18', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DATA17', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DATA16', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DATA15', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DATA14', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DATA13', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DATA12', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DATA11', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DATA10', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DATA9', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DATA8', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DATA7', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DATA6', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DATA5', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DATA4', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DATA3', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DATA2', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DATA1', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DATA0', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DIR31', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DIR30', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DIR29', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DIR28', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DIR27', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DIR26', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DIR25', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DIR24', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DIR23', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DIR22', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DIR21', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DIR20', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DIR19', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DIR18', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DIR17', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DIR16', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DIR15', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DIR14', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DIR13', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DIR12', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DIR11', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DIR10', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DIR9', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DIR8', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DIR7', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DIR6', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DIR5', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DIR4', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DIR3', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DIR2', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DIR1', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_DIR0', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:DRIVE_FREQ', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:DRIVE_AMPL', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:DRIVE_SPAN', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:DRIVE_PERIOD', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:DRIVE_WAVEFORM', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:DRIVE_PATTERN', _EMPTY_STR, 0.0],
    ['SI-Glob:DI-BbBProc-H:DRIVE_MASK', _EMPTY_NDARRAY_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:FE_PHASE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:FE_ATTEN', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:BE_PHASE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:BE_ATTEN', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:TADC', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:TDAC', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:OFF_FIDC', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:OFF_FIDS', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:OFF_DAC', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:DEL_CAL', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:PHASE_SERVO_OFFSET', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:PHASE_SERVO_SETPT', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:PHASE_SERVO_MAXDELTA', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:PHASE_SERVO_GAIN', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:PHASE_SERVO_MODE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:PHASE_SERVO_SIGN', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:GPIO_SEL', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:SAT_THRESHOLD', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:FB_PATTERN', _EMPTY_STR, 0.0],
    ['SI-Glob:DI-BbBProc-H:FB_MASK', _EMPTY_NDARRAY_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:CLEAN_TUNE', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:CLEAN_AMPL', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:CLEAN_PATTERN', _EMPTY_STR, 0.0],
    ['SI-Glob:DI-BbBProc-H:CLEAN_ENABLE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:CLEAN_RENABLE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:CLEAN_SAVE_FREQ', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:CLEAN_SAVE_AMPL', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:CLEAN_SAVE_DRV_PATTERN', _EMPTY_STR, 0.0],
    ['SI-Glob:DI-BbBProc-H:CLEAN_SAVE_FB_PATTERN', _EMPTY_STR, 0.0],
    ['SI-Glob:DI-BbBProc-H:CLEAN_SAVE_MODE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:MMGRAW_0_SLOPE', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:MMGRAW_1_SLOPE', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:MMGRAW_0_OFFSET', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:MMGRAW_1_OFFSET', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:MCLRAW_0_FWDLOSS', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:MCLRAW_1_FWDLOSS', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:MCLRAW_0_REVLOSS', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:MCLRAW_1_REVLOSS', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:LEVEL_FID_ENABLE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:LEVEL_TRIG1_ENABLE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:LEVEL_TRIG2_ENABLE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:AD5644_V_CH0', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:AD5644_V_CH1', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:AD5644_V_CH2', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:AD5644_V_CH3', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:AD5644_V_CH4', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:AD5644_V_CH5', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:AD5644_V_CH6', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:AD5644_V_CH7', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:AD5644_V_CH8', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:AD5644_V_CH9', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:AD5644_V_CH10', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:AD5644_V_CH11', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:AD5644REF0_BO', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:AD5644REF1_BO', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:AD5644REF2_BO', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:LEVEL_FID', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:LEVEL_TRIG1', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:LEVEL_TRIG2', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:LEVEL_VTRIG1', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:LEVEL_VTRIG2', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:PHPLL', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:FBELT_SERVO_SETPT', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:FBE_X_PHASE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:FBE_Y_PHASE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:FBE_BE_PHASE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:FBE_Z_ATT', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:FBE_X_ATT', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:FBE_Y_ATT', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:FBE_BE_ATT', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:FBELT_SERVO_MODE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:FBELT_SERVO_SIGN', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:FBELT_SERVO_GAIN', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:FBELT_SERVO_OFFSET', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:FBELT_SERVO_MAXDELTA', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:FBELT_FAN_SETPT', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:FBELT_FAN_MODE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:FBELT_FAN_GAIN_P', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:FBELT_FAN_GAIN_I', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:SHAPE_C0', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:SHAPE_C2', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:SRAM_DACQ', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:SRAM_ACQTIME', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:SRAM_GDTIME', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:SRAM_HOLDTIME', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:SRAM_POSTTIME', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:SRAM_ARM', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:SRAM_EXTEN', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:SRAM_REC_DS', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:SRAM_BR_ARM', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:SRAM_ACQ_SINGLE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:SRAM_ACQ_EN', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:SRAM_ACQ_PATTERN', _EMPTY_STR, 0.0],
    ['SI-Glob:DI-BbBProc-H:SRAM_ACQ_MASK', _EMPTY_NDARRAY_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:SRAM_SP_LOW1', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:SRAM_SP_HIGH1', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:SRAM_SP_SEARCH1', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:SRAM_SP_LOW2', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:SRAM_SP_HIGH2', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:SRAM_SP_SEARCH2', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:SRAM_SP_AVG', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:SRAM_TRIG_IN_SEL', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:SRAM_POSTSEL', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:BRAM_DACQ', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:BRAM_ACQTIME', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:BRAM_GDTIME', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:BRAM_HOLDTIME', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:BRAM_POSTTIME', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:BRAM_ARM', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:BRAM_EXTEN', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:BRAM_REC_DS', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:BRAM_BR_ARM', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:BRAM_ACQ_SINGLE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:BRAM_ACQ_EN', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:BRAM_ACQ_PATTERN', _EMPTY_STR, 0.0],
    ['SI-Glob:DI-BbBProc-H:BRAM_ACQ_MASK', _EMPTY_NDARRAY_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:BRAM_SP_LOW1', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:BRAM_SP_HIGH1', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:BRAM_SP_SEARCH1', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:BRAM_SP_LOW2', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:BRAM_SP_HIGH2', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:BRAM_SP_SEARCH2', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:BRAM_SP_AVG', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-H:BRAM_TRIG_IN_SEL', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:BRAM_POSTSEL', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-H:FBE_DEV_ID', _EMPTY_STR, 0.0]]


_pvs_bbb_proc_v = [
    ['SI-Glob:DI-BbBProc-V:GDEN', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:FBCTRL', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:SETSEL', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:DELAY', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:CLKRST', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:SHIFTGAIN', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:PROC_DS', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:COEFF', _EMPTY_NDARRAY_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:CSET0', _EMPTY_NDARRAY_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:CSET1', _EMPTY_NDARRAY_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:LDSET', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:DESC_COEFF', _EMPTY_STR, 0.0],
    ['SI-Glob:DI-BbBProc-V:DESC_CSET0', _EMPTY_STR, 0.0],
    ['SI-Glob:DI-BbBProc-V:DESC_CSET1', _EMPTY_STR, 0.0],
    ['SI-Glob:DI-BbBProc-V:FID_DELAY', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:FLT_GAIN', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:FLT_PHASE', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:FLT_FREQ', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:FLT_TAPS', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:FTF_TUNE', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DATA31', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DATA30', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DATA29', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DATA28', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DATA27', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DATA26', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DATA25', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DATA24', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DATA23', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DATA22', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DATA21', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DATA20', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DATA19', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DATA18', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DATA17', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DATA16', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DATA15', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DATA14', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DATA13', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DATA12', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DATA11', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DATA10', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DATA9', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DATA8', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DATA7', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DATA6', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DATA5', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DATA4', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DATA3', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DATA2', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DATA1', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DATA0', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DIR31', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DIR30', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DIR29', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DIR28', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DIR27', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DIR26', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DIR25', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DIR24', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DIR23', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DIR22', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DIR21', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DIR20', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DIR19', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DIR18', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DIR17', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DIR16', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DIR15', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DIR14', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DIR13', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DIR12', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DIR11', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DIR10', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DIR9', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DIR8', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DIR7', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DIR6', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DIR5', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DIR4', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DIR3', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DIR2', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DIR1', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_DIR0', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:DRIVE_FREQ', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:DRIVE_AMPL', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:DRIVE_SPAN', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:DRIVE_PERIOD', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:DRIVE_WAVEFORM', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:DRIVE_PATTERN', _EMPTY_STR, 0.0],
    ['SI-Glob:DI-BbBProc-V:DRIVE_MASK', _EMPTY_NDARRAY_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:FE_PHASE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:FE_ATTEN', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:BE_PHASE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:BE_ATTEN', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:TADC', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:TDAC', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:OFF_FIDC', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:OFF_FIDS', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:OFF_DAC', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:DEL_CAL', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:PHASE_SERVO_OFFSET', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:PHASE_SERVO_SETPT', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:PHASE_SERVO_MAXDELTA', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:PHASE_SERVO_GAIN', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:PHASE_SERVO_MODE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:PHASE_SERVO_SIGN', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:GPIO_SEL', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:SAT_THRESHOLD', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:FB_PATTERN', _EMPTY_STR, 0.0],
    ['SI-Glob:DI-BbBProc-V:FB_MASK', _EMPTY_NDARRAY_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:CLEAN_TUNE', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:CLEAN_AMPL', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:CLEAN_PATTERN', _EMPTY_STR, 0.0],
    ['SI-Glob:DI-BbBProc-V:CLEAN_ENABLE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:CLEAN_SAVE_FREQ', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:CLEAN_SAVE_AMPL', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:CLEAN_SAVE_DRV_PATTERN', _EMPTY_STR, 0.0],
    ['SI-Glob:DI-BbBProc-V:CLEAN_SAVE_FB_PATTERN', _EMPTY_STR, 0.0],
    ['SI-Glob:DI-BbBProc-V:CLEAN_SAVE_MODE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:MMGRAW_0_SLOPE', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:MMGRAW_1_SLOPE', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:MMGRAW_0_OFFSET', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:MMGRAW_1_OFFSET', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:MCLRAW_0_FWDLOSS', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:MCLRAW_1_FWDLOSS', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:MCLRAW_0_REVLOSS', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:MCLRAW_1_REVLOSS', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:LEVEL_FID_ENABLE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:LEVEL_TRIG1_ENABLE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:LEVEL_TRIG2_ENABLE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:AD5644_V_CH0', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:AD5644_V_CH1', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:AD5644_V_CH2', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:AD5644_V_CH3', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:AD5644_V_CH4', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:AD5644_V_CH5', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:AD5644_V_CH6', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:AD5644_V_CH7', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:AD5644_V_CH8', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:AD5644_V_CH9', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:AD5644_V_CH10', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:AD5644_V_CH11', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:AD5644REF0_BO', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:AD5644REF1_BO', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:AD5644REF2_BO', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:LEVEL_FID', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:LEVEL_TRIG1', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:LEVEL_TRIG2', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:LEVEL_VTRIG1', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:LEVEL_VTRIG2', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:PHPLL', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:FBELT_SERVO_SETPT', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:FBE_X_PHASE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:FBE_Y_PHASE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:FBE_BE_PHASE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:FBE_Z_ATT', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:FBE_X_ATT', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:FBE_Y_ATT', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:FBE_BE_ATT', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:FBELT_SERVO_MODE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:FBELT_SERVO_SIGN', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:FBELT_SERVO_GAIN', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:FBELT_SERVO_OFFSET', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:FBELT_SERVO_MAXDELTA', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:FBELT_FAN_SETPT', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:FBELT_FAN_MODE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:FBELT_FAN_GAIN_P', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:FBELT_FAN_GAIN_I', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:SHAPE_C0', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:SHAPE_C2', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:SRAM_DACQ', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:SRAM_ACQTIME', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:SRAM_GDTIME', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:SRAM_HOLDTIME', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:SRAM_POSTTIME', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:SRAM_ARM', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:SRAM_EXTEN', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:SRAM_REC_DS', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:SRAM_BR_ARM', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:SRAM_ACQ_SINGLE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:SRAM_ACQ_EN', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:SRAM_ACQ_PATTERN', _EMPTY_STR, 0.0],
    ['SI-Glob:DI-BbBProc-V:SRAM_ACQ_MASK', _EMPTY_NDARRAY_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:SRAM_SP_LOW1', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:SRAM_SP_HIGH1', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:SRAM_SP_SEARCH1', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:SRAM_SP_LOW2', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:SRAM_SP_HIGH2', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:SRAM_SP_SEARCH2', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:SRAM_SP_AVG', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:SRAM_TRIG_IN_SEL', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:SRAM_POSTSEL', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:BRAM_DACQ', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:BRAM_ACQTIME', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:BRAM_GDTIME', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:BRAM_HOLDTIME', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:BRAM_POSTTIME', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:BRAM_ARM', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:BRAM_EXTEN', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:BRAM_REC_DS', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:BRAM_BR_ARM', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:BRAM_ACQ_SINGLE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:BRAM_ACQ_EN', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:BRAM_ACQ_PATTERN', _EMPTY_STR, 0.0],
    ['SI-Glob:DI-BbBProc-V:BRAM_ACQ_MASK', _EMPTY_NDARRAY_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:BRAM_SP_LOW1', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:BRAM_SP_HIGH1', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:BRAM_SP_SEARCH1', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:BRAM_SP_LOW2', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:BRAM_SP_HIGH2', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:BRAM_SP_SEARCH2', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:BRAM_SP_AVG', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-V:BRAM_TRIG_IN_SEL', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:BRAM_POSTSEL', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-V:FBE_DEV_ID', _EMPTY_STR, 0.0]]


_pvs_bbb_proc_l = [
    ['SI-Glob:DI-BbBProc-L:GDEN', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:FBCTRL', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:SETSEL', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:DELAY', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:CLKRST', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:SHIFTGAIN', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:PROC_DS', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:COEFF', _EMPTY_NDARRAY_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:CSET0', _EMPTY_NDARRAY_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:CSET1', _EMPTY_NDARRAY_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:LDSET', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:DESC_COEFF', _EMPTY_STR, 0.0],
    ['SI-Glob:DI-BbBProc-L:DESC_CSET0', _EMPTY_STR, 0.0],
    ['SI-Glob:DI-BbBProc-L:DESC_CSET1', _EMPTY_STR, 0.0],
    ['SI-Glob:DI-BbBProc-L:FID_DELAY', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:FLT_GAIN', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:FLT_PHASE', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:FLT_FREQ', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:FLT_TAPS', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:FTF_TUNE', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DATA31', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DATA30', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DATA29', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DATA28', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DATA27', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DATA26', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DATA25', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DATA24', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DATA23', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DATA22', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DATA21', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DATA20', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DATA19', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DATA18', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DATA17', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DATA16', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DATA15', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DATA14', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DATA13', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DATA12', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DATA11', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DATA10', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DATA9', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DATA8', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DATA7', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DATA6', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DATA5', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DATA4', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DATA3', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DATA2', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DATA1', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DATA0', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DIR31', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DIR30', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DIR29', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DIR28', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DIR27', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DIR26', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DIR25', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DIR24', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DIR23', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DIR22', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DIR21', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DIR20', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DIR19', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DIR18', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DIR17', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DIR16', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DIR15', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DIR14', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DIR13', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DIR12', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DIR11', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DIR10', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DIR9', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DIR8', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DIR7', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DIR6', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DIR5', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DIR4', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DIR3', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DIR2', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DIR1', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_DIR0', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:DRIVE_FREQ', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:DRIVE_AMPL', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:DRIVE_SPAN', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:DRIVE_PERIOD', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:DRIVE_WAVEFORM', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:DRIVE_PATTERN', _EMPTY_STR, 0.0],
    ['SI-Glob:DI-BbBProc-L:DRIVE_MASK', _EMPTY_NDARRAY_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:FE_PHASE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:FE_ATTEN', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:BE_PHASE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:BE_ATTEN', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:TADC', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:TDAC', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:OFF_FIDC', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:OFF_FIDS', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:OFF_DAC', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:DEL_CAL', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:PHASE_SERVO_OFFSET', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:PHASE_SERVO_SETPT', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:PHASE_SERVO_MAXDELTA', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:PHASE_SERVO_GAIN', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:PHASE_SERVO_MODE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:PHASE_SERVO_SIGN', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:GPIO_SEL', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:SAT_THRESHOLD', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:FB_PATTERN', _EMPTY_STR, 0.0],
    ['SI-Glob:DI-BbBProc-L:FB_MASK', _EMPTY_NDARRAY_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:CLEAN_TUNE', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:CLEAN_AMPL', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:CLEAN_PATTERN', _EMPTY_STR, 0.0],
    ['SI-Glob:DI-BbBProc-L:CLEAN_ENABLE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:CLEAN_SAVE_FREQ', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:CLEAN_SAVE_AMPL', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:CLEAN_SAVE_DRV_PATTERN', _EMPTY_STR, 0.0],
    ['SI-Glob:DI-BbBProc-L:CLEAN_SAVE_FB_PATTERN', _EMPTY_STR, 0.0],
    ['SI-Glob:DI-BbBProc-L:CLEAN_SAVE_MODE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:MMGRAW_0_SLOPE', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:MMGRAW_1_SLOPE', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:MMGRAW_0_OFFSET', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:MMGRAW_1_OFFSET', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:MCLRAW_0_FWDLOSS', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:MCLRAW_1_FWDLOSS', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:MCLRAW_0_REVLOSS', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:MCLRAW_1_REVLOSS', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:LEVEL_FID_ENABLE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:LEVEL_TRIG1_ENABLE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:LEVEL_TRIG2_ENABLE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:AD5644_V_CH0', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:AD5644_V_CH1', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:AD5644_V_CH2', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:AD5644_V_CH3', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:AD5644_V_CH4', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:AD5644_V_CH5', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:AD5644_V_CH6', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:AD5644_V_CH7', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:AD5644_V_CH8', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:AD5644_V_CH9', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:AD5644_V_CH10', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:AD5644_V_CH11', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:AD5644REF0_BO', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:AD5644REF1_BO', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:AD5644REF2_BO', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:LEVEL_FID', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:LEVEL_TRIG1', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:LEVEL_TRIG2', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:LEVEL_VTRIG1', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:LEVEL_VTRIG2', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:PHPLL', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:FBELT_SERVO_SETPT', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:FBE_X_PHASE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:FBE_Y_PHASE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:FBE_BE_PHASE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:FBE_Z_ATT', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:FBE_X_ATT', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:FBE_Y_ATT', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:FBE_BE_ATT', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:FBELT_SERVO_MODE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:FBELT_SERVO_SIGN', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:FBELT_SERVO_GAIN', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:FBELT_SERVO_OFFSET', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:FBELT_SERVO_MAXDELTA', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:FBELT_FAN_SETPT', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:FBELT_FAN_MODE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:FBELT_FAN_GAIN_P', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:FBELT_FAN_GAIN_I', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:SHAPE_C0', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:SHAPE_C2', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:SRAM_DACQ', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:SRAM_ACQTIME', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:SRAM_GDTIME', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:SRAM_HOLDTIME', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:SRAM_POSTTIME', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:SRAM_ARM', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:SRAM_EXTEN', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:SRAM_REC_DS', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:SRAM_BR_ARM', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:SRAM_ACQ_SINGLE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:SRAM_ACQ_EN', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:SRAM_ACQ_PATTERN', _EMPTY_STR, 0.0],
    ['SI-Glob:DI-BbBProc-L:SRAM_ACQ_MASK', _EMPTY_NDARRAY_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:SRAM_SP_LOW1', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:SRAM_SP_HIGH1', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:SRAM_SP_SEARCH1', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:SRAM_SP_LOW2', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:SRAM_SP_HIGH2', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:SRAM_SP_SEARCH2', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:SRAM_SP_AVG', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:SRAM_TRIG_IN_SEL', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:SRAM_POSTSEL', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:BRAM_DACQ', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:BRAM_ACQTIME', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:BRAM_GDTIME', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:BRAM_HOLDTIME', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:BRAM_POSTTIME', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:BRAM_ARM', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:BRAM_EXTEN', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:BRAM_REC_DS', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:BRAM_BR_ARM', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:BRAM_ACQ_SINGLE', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:BRAM_ACQ_EN', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:BRAM_ACQ_PATTERN', _EMPTY_STR, 0.0],
    ['SI-Glob:DI-BbBProc-L:BRAM_ACQ_MASK', _EMPTY_NDARRAY_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:BRAM_SP_LOW1', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:BRAM_SP_HIGH1', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:BRAM_SP_SEARCH1', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:BRAM_SP_LOW2', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:BRAM_SP_HIGH2', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:BRAM_SP_SEARCH2', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:BRAM_SP_AVG', _ZERO_FLOAT, 0.0],
    ['SI-Glob:DI-BbBProc-L:BRAM_TRIG_IN_SEL', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:BRAM_POSTSEL', _ZERO_INT, 0.0],
    ['SI-Glob:DI-BbBProc-L:FBE_DEV_ID', _EMPTY_STR, 0.0]]


_template_dict = {
    'pvs':
        _pvs_bbb_proc_h + _pvs_bbb_proc_v + _pvs_bbb_proc_l
    }



"""."""

import time as _time
import numpy as _np

from .device import DeviceNC as _DeviceNC
from .device import Devices as _Devices


class RFGen(_DeviceNC):
    """."""

    RF_DELTA_MIN = 0.1  # [Hz]
    RF_DELTA_MAX = 15000.0  # [Hz]
    RF_DELTA_RMP = 20  # [Hz]

    class DEVICES:
        """Devices names."""

        AS = 'RF-Gen'
        ALL = (AS, )

    _properties = ('GeneralFreq-SP', 'GeneralFreq-RB')

    def __init__(self, devname=None):
        """."""
        if devname is None:
            devname = RFGen.DEVICES.AS

        # check if device exists
        if devname not in RFGen.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        super().__init__(devname, properties=RFGen._properties)

    @property
    def frequency(self):
        """."""
        return self['GeneralFreq-RB']

    @frequency.setter
    def frequency(self, value):
        delta_max = RFGen.RF_DELTA_RMP  # [Hz]
        freq0 = self['GeneralFreq-SP']
        if freq0 is None or value is None:
            return
        delta = abs(value-freq0)
        if delta < RFGen.RF_DELTA_MIN or delta > RFGen.RF_DELTA_MAX:
            return
        npoints = int(round(delta/delta_max)) + 2
        freq_span = _np.linspace(freq0, value, npoints)[1:]
        for freq in freq_span:
            self._pvs['GeneralFreq-SP'].put(freq, wait=False)
            _time.sleep(1.0)
        self['GeneralFreq-SP'] = value


class ASLLRF(_DeviceNC):
    """."""

    class DEVICES:
        """Devices names."""

        BO = 'BR-RF-DLLRF-01'
        SI = 'SR-RF-DLLRF-01'
        ALL = (BO, SI)

    _properties = (
        'PL:REF:S', 'SL:REF:PHS', 'SL:INP:PHS',
        'mV:AL:REF:S', 'SL:REF:AMP', 'SL:INP:AMP',
        'DTune-SP', 'DTune-RB', 'TUNE:DEPHS',
        'RmpPhsBot-SP', 'RmpPhsBot-RB', 'RmpPhsTop-SP', 'RmpPhsTop-RB',
        'RmpEnbl-Sts',
        )

    def __init__(self, devname):
        """."""
        # check if device exists
        if devname not in ASLLRF.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        super().__init__(devname, properties=ASLLRF._properties)

    @property
    def is_cw(self):
        """."""
        return not self['RmpEnbl-Sts']

    @property
    def phase_top(self):
        """."""
        return self['RmpPhsTop-RB']

    @phase_top.setter
    def phase_top(self, value):
        """."""
        self['RmpPhsTop-SP'] = value

    @property
    def phase_bottom(self):
        """."""
        return self['RmpPhsBot-RB']

    @phase_bottom.setter
    def phase_bottom(self, value):
        """."""
        self['RmpPhsBot-SP'] = value

    @property
    def phase_mon(self):
        """."""
        return self['SL:INP:PHS']

    @property
    def phase(self):
        """."""
        return self['SL:REF:PHS']

    @phase.setter
    def phase(self, value):
        """."""
        self['PL:REF:S'] = value

    @property
    def voltage_mon(self):
        """."""
        return self['SL:INP:AMP']

    @property
    def voltage(self):
        """."""
        return self['SL:REF:AMP']

    @voltage.setter
    def voltage(self, value):
        self['mV:AL:REF:S'] = value

    @property
    def detune(self):
        """."""
        return self['DTune-RB']

    @detune.setter
    def detune(self, value):
        """."""
        self['DTune-SP'] = value

    @property
    def detune_error(self):
        """."""
        return self['TUNE:DEPHS']


class BORFCavMonitor(_DeviceNC):
    """."""

    class DEVICES:
        """Devices names."""

        BO = 'BO-05D:RF-P5Cav'

    _properties = (
        'Cell3PwrTop-Mon', 'Cell3PwrBot-Mon', 'PwrRFIntlk-Mon', 'Sts-Mon',
        'Cell1Pwr-Mon', 'Cell2Pwr-Mon', 'Cell3Pwr-Mon', 'Cell4Pwr-Mon',
        'Cell5Pwr-Mon', 'Cylin1T-Mon', 'Cylin2T-Mon', 'Cylin3T-Mon',
        'Cylin4T-Mon', 'Cylin5T-Mon', 'CoupT-Mon',
        )

    def __init__(self):
        """."""
        # call base class constructor
        super().__init__(
            BORFCavMonitor.DEVICES.BO, properties=BORFCavMonitor._properties)

    @property
    def status(self):
        """."""
        return self['Sts-Mon']

    @property
    def power_interlock(self):
        """."""
        return self['PwrRFIntlk-Mon']

    @property
    def power_top(self):
        """."""
        return self['Cell3PwrTop-Mon']

    @property
    def power_bottom(self):
        """."""
        return self['Cell3PwrBot-Mon']

    @property
    def power_reverse(self):
        """."""
        return self['PwrRev-Mon']

    @property
    def power_forward(self):
        """."""
        return self['PwrFwd-Mon']

    @property
    def power_cell1(self):
        """."""
        return self['Cell1Pwr-Mon']

    @property
    def power_cell2(self):
        """."""
        return self['Cell2Pwr-Mon']

    @property
    def power_cell3(self):
        """."""
        return self['Cell3Pwr-Mon']

    @property
    def power_cell4(self):
        """."""
        return self['Cell4Pwr-Mon']

    @property
    def power_cell5(self):
        """."""
        return self['Cell5Pwr-Mon']

    @property
    def temp_coupler(self):
        """."""
        return self['CoupT-Mon']

    @property
    def temp_cell1(self):
        """."""
        return self['Cylin1T-Mon']

    @property
    def temp_cell2(self):
        """."""
        return self['Cylin2T-Mon']

    @property
    def temp_cell3(self):
        """."""
        return self['Cylin3T-Mon']

    @property
    def temp_cell4(self):
        """."""
        return self['Cylin4T-Mon']

    @property
    def temp_cell5(self):
        """."""
        return self['Cylin5T-Mon']


class SIRFCavMonitor(_DeviceNC):
    """."""

    class DEVICES:
        """Devices names."""

        SI = 'SI-02SB:RF-P7Cav'

    _properties = (
        'PwrCell4Top-Mon', 'PwrCell4Bot-Mon', 'PwrRFIntlk-Mon', 'Sts-Mon',
        'PwrCell2-Mon', 'PwrCell4-Mon', 'PwrCell6-Mon', 'Cylin1T-Mon',
        'Cylin2T-Mon', 'Cylin3T-Mon', 'Cylin4T-Mon', 'Cylin5T-Mon',
        'Cylin6T-Mon', 'Cylin7T-Mon', 'CoupT-Mon'
        )

    def __init__(self):
        """."""
        # call base class constructor
        super().__init__(
            SIRFCavMonitor.DEVICES.SI, properties=SIRFCavMonitor._properties)

    @property
    def status(self):
        """."""
        return self['Sts-Mon']

    @property
    def power_interlock(self):
        """."""
        return self['PwrRFIntlk-Mon']

    @property
    def power_top(self):
        """."""
        return self['PwrCell4Top-Mon']

    @property
    def power_bottom(self):
        """."""
        return self['PwrCell4Bot-Mon']

    @property
    def power_reverse(self):
        """."""
        return self['PwrRev-Mon']

    @property
    def power_forward(self):
        """."""
        return self['PwrFwd-Mon']

    @property
    def power_cell2(self):
        """."""
        return self['PwrCell2-Mon']

    @property
    def power_cell4(self):
        """."""
        return self['PwrCell4-Mon']

    @property
    def power_cell6(self):
        """."""
        return self['PwrCell6-Mon']

    @property
    def temp_coupler(self):
        """."""
        return self['CoupT-Mon']

    @property
    def temp_cell1(self):
        """."""
        return self['Cylin1T-Mon']

    @property
    def temp_cell2(self):
        """."""
        return self['Cylin2T-Mon']

    @property
    def temp_cell3(self):
        """."""
        return self['Cylin3T-Mon']

    @property
    def temp_cell4(self):
        """."""
        return self['Cylin4T-Mon']

    @property
    def temp_cell5(self):
        """."""
        return self['Cylin5T-Mon']

    @property
    def temp_cell6(self):
        """."""
        return self['Cylin6T-Mon']

    @property
    def temp_cell7(self):
        """."""
        return self['Cylin7T-Mon']


class RFCav(_Devices):
    """."""

    class DEVICES:
        """Devices names."""

        BO = 'BO-05D:RF-P5Cav'
        SI = 'SI-02SB:RF-P7Cav'
        ALL = (BO, SI)

    def __init__(self, devname):
        """."""
        # check if device exists
        if devname not in RFCav.DEVICES.ALL:
            raise NotImplementedError(devname)

        rfgen = RFGen()
        if devname == RFCav.DEVICES.SI:
            llrf = ASLLRF(ASLLRF.DEVICES.SI)
            rfmon = SIRFCavMonitor()
        elif devname == RFCav.DEVICES.BO:
            llrf = ASLLRF(ASLLRF.DEVICES.BO)
            rfmon = BORFCavMonitor()
        devices = (rfgen, llrf, rfmon)

        # call base class constructor
        super().__init__(devname, devices)

    @property
    def is_cw(self):
        """."""
        return self.dev_llrf.is_cw

    @property
    def power(self):
        """."""
        return self.dev_cavmon.get_power(self.is_cw)

    @property
    def dev_rfgen(self):
        """Return RFGen device."""
        return self.devices[0]

    @property
    def dev_llrf(self):
        """Return LLRF device."""
        return self.devices[1]

    @property
    def dev_cavmon(self):
        """Return RFPoweMon device."""
        return self.devices[2]

    def cmd_set_voltage(self, value, timeout=10):
        """."""
        self.dev_llrf.voltage = value
        self._wait('voltage', timeout=timeout)

    def cmd_set_phase(self, value, timeout=10):
        """."""
        self.dev_llrf.phase = value
        self._wait('phase', timeout=timeout)

    def cmd_set_frequency(self, value, timeout=10):
        """."""
        self.dev_rfgen.frequency = value
        self._wait('frequency', timeout=timeout)

    # --- private methods ---

    def _wait(self, propty, timeout=10):
        """."""
        nrp = int(timeout / 0.1)
        for _ in range(nrp):
            _time.sleep(0.1)
            if propty == 'phase':
                if self.is_cw:
                    phase_sp = self.dev_llrf['PL:REF:S']
                else:
                    phase_sp = self.dev_llrf['RmpPhsBot-SP']
                if abs(self.dev_llrf.phase - phase_sp) < 0.1:
                    break
            elif propty == 'voltage':
                voltage_sp = self.dev_llrf['mV:AL:REF:S']
                if abs(self.dev_llrf.voltage - voltage_sp) < 0.1:
                    break
            elif propty == 'frequency':
                freq_sp = self.dev_rfgen['GeneralFreq-SP']
                if abs(self.dev_rfgen.frequency - freq_sp) < 0.1:
                    break
            else:
                raise Exception(
                    'Set RF property (phase, voltage or frequency)')



"""Devices subpackage."""

from .device import Device, DeviceApp, Devices
from .bpm import BPM
from .dcct import DCCT
from .egun import EGBias, EGFilament, EGHVPS
from .ict import ICT, TranspEff
from .lillrf import LILLRF
from .orbit_interlock import BPMInterlock
from .pwrsupply import PowerSupply, PowerSupplyPU
from .psconv import PSProperty, StrengthConv
from .pssofb import PSCorrSOFB, PSApplySOFB
from .rf import RFGen, ASLLRF, BORFCavMonitor, SIRFCavMonitor, RFCav
from .screen import Screen
from .tune import TuneFrac, TuneProc, Tune, TuneCorr
from .sofb import SOFB
from .timing import EVG, Event
from .syncd import DevicesSync
from .energy import Energy
from .ids import IDCorrectors, APU, APUFeedForward
from .currinfo import CurrInfoTranspEff, CurrInfoLinear, \
    CurrInfoBO, CurrInfoSI, CurrInfoAS
from .blscreen import BeamlineScreen
from .bbb import BunchbyBunch

del device, bpm, dcct, egun, ict, lillrf,
del pwrsupply, psconv, pssofb, rf,
del screen, tune, sofb, timing, syncd, energy
del ids, currinfo, blscreen, bbb
